
// Copyright (C) Tom Horsley, 2001,2004. All rights reserved.

// Control panel applet for NTPTime.

#include <windows.h>
#include <commctrl.h>
#include <cpl.h>
#include <shlobj.h>
#include "resource.h"

#pragma warning(disable:4786)

#include <string>
#include <vector>

// All the registry values for NTPTime and flags indicating if they were
// found in the registry (and what format they were).

bool changed = false;

static char * NTPServerName = NULL;

static DWORD NTPServerName_fmt = -1;

static char * LogDirectory = NULL;

static DWORD LogDirectory_fmt = REG_SZ;

static DWORD WriteLogEntries = 0;

static DWORD WriteLogEntries_fmt = -1;

static DWORD LogFileCount = 0;

static DWORD LogFileCount_fmt = -1;

static bool  use_LogFileCount = false;

static DWORD NTPServerPort = 123;

static DWORD NTPServerPort_fmt = -1;

static bool  use_NTPServerPort = false;

static DWORD WaitForRAS = 0;

static DWORD WaitForRAS_fmt = -1;

static DWORD RASPollMilliseconds = 0;

static DWORD RASPollMilliseconds_fmt = -1;

static DWORD NTPPollMilliseconds = 0;

static DWORD NTPPollMilliseconds_fmt = -1;

static DWORD RunOnce = 1;

static DWORD RunOnce_fmt = -1;

static DWORD MaxAdjustSeconds = 0;

static DWORD MaxAdjustSeconds_fmt = -1;

static bool  use_MaxAdjustSeconds = false;

static DWORD AllowPoke = 1;

static DWORD AllowPoke_fmt = -1;

static DWORD AllowMonitor = 1;

static DWORD AllowMonitor_fmt = -1;

static DWORD RetryMaxCount = 0;

static DWORD RetryMaxCount_fmt = -1;

static bool  use_RetryMaxCount = false;

static DWORD RetryMinMilliseconds = 60*1000;

static DWORD RetryMinMilliseconds_fmt = -1;

static DWORD RetryMaxMilliseconds = 6*60*60*1000;

static DWORD RetryMaxMilliseconds_fmt = -1;

static DWORD RetryAdd = 0;

static DWORD RetryAdd_fmt = -1;

static DWORD RetryMult = 2;

static DWORD RetryMult_fmt = -1;

static DWORD RetryDiv = 1;

static DWORD RetryDiv_fmt = -1;

static DWORD NtpProtocolVersion = 3;

static DWORD NtpProtocolVersion_fmt = -1;

static bool  use_NtpProtocolVersion = false;

static DWORD StartupDelayMilliseconds = 0;

static DWORD StartupDelayMilliseconds_fmt = -1;

static DWORD RasDelayMilliseconds = 0;

static DWORD RasDelayMilliseconds_fmt = -1;

static DWORD ShutdownDelayMilliseconds = 0;

static DWORD ShutdownDelayMilliseconds_fmt = -1;

static DWORD DisablePacketErrorChecks = 0;

static DWORD DisablePacketErrorChecks_fmt = -1;

static bool  use_NtpReplyPort = false;

static DWORD NtpReplyPort = 0;

static DWORD NtpReplyPort_fmt = -1;

// Possible previous versions of NTPTime registry to look in for values
// NOTE - Need to add new version at the start of this list each time
// a new release of ntptime comes out.
//
static char * versions[] = { "4.3", "4.2", "4.1", "4.0", "3.1", "3.0",
                             "2.3", "2.2", "2.1" };

static HINSTANCE hinst = NULL;

static std::vector<std::string> * server_name_list_ptr = NULL;

static int selected_server = -1;

static std::vector<std::string> &
GetServerNameList() {
   if (server_name_list_ptr == NULL) {
      server_name_list_ptr = new std::vector<std::string>;
   }
   return *server_name_list_ptr;
}

static void
AppendServerNames(const char * str) {
   std::string one_server;
   char c;
   bool skipping = true;
   while ((c = *str++) != '\0') {
      if (skipping) {
         if (! ((c == ' ') || (c == ',') || (c == ';'))) {
            skipping = false;
            one_server.append(1, c);
         }
      } else {
         if ((c == ' ') || (c == ',') || (c == ';')) {
            GetServerNameList().push_back(one_server);
            one_server.erase();
            skipping = true;
         } else {
            one_server.append(1, c);
         }
      }
   }
   if (one_server.size() > 0) {
      GetServerNameList().push_back(one_server);
   }
}

// GetRegString fetches a string value from the registry and records
// the format (REG_SZ or REG_EXPAND_SZ).
//
static void
GetRegString(HKEY k, const char * vname, char *& valptr, DWORD & fmt) {
   char * rval = NULL;
   DWORD len = 0;
   if ((RegQueryValueEx(k, vname, 0, &fmt, NULL, &len) == ERROR_SUCCESS) &&
       ((fmt == REG_SZ) || (fmt == REG_EXPAND_SZ))) {
      rval = new char [len];
      if ((RegQueryValueEx(k, vname, 0, &fmt, (LPBYTE)rval, &len) !=
           ERROR_SUCCESS) ||
          ((fmt != REG_SZ) && (fmt != REG_EXPAND_SZ))) {
         delete [] rval;
         rval = NULL;
      }
   }
   if (rval != NULL) {
      delete [] valptr;
      valptr = rval;
   }
}

// GetRegDWORD fetches a DWORD value from the registry and records the
// format (should always be REG_DWORD).
//
static void
GetRegDWORD(HKEY k, const char * vname, DWORD& val, DWORD & fmt) {
   DWORD rval;
   DWORD len = sizeof(rval);
   if ((RegQueryValueEx(k, vname, 0, &fmt, (LPBYTE)&rval, &len) ==
        ERROR_SUCCESS) &&
       (fmt == REG_DWORD)) {
      val = rval;
   }
}

// ReadNTPTimeRegistry looks for existing NTPTime registry keys, starting
// with the newest version number and working down to the oldest. It stops
// as soon as it finds a key that exists, and reads the values into the
// global variables (declared above).
//
static void
ReadNTPTimeRegistry() {
   HKEY k;
   char keyname[75];
   strcpy(keyname, "SOFTWARE\\ntcli\\NTPTime\\");
   int keylen = strlen(keyname);
   for (int i = 0; i < sizeof(versions)/sizeof(versions[0]); ++i) {
      strcpy(&keyname[keylen], versions[i]);
      if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, keyname, 0, KEY_READ, &k) ==
             ERROR_SUCCESS) {
         if (i != 0) {

            // If we didn't get these values from the current version
            // registry key, need to pretend they have changed in order to
            // make sure they get written to the current key.

            changed = true;
         }
         GetRegString(k, "NTPServerName",
            NTPServerName, NTPServerName_fmt);
         GetRegString(k, "LogDirectory",
            LogDirectory, LogDirectory_fmt);
         GetRegDWORD(k, "WriteLogEntries",
            WriteLogEntries, WriteLogEntries_fmt);
         GetRegDWORD(k, "LogFileCount",
            LogFileCount, LogFileCount_fmt);
         use_LogFileCount = (LogFileCount >= 2);
         GetRegDWORD(k, "NTPServerPort",
            NTPServerPort, NTPServerPort_fmt);
         use_NTPServerPort = (NTPServerPort != 123);
         GetRegDWORD(k, "WaitForRAS",
            WaitForRAS, WaitForRAS_fmt);
         GetRegDWORD(k, "RASPollMilliseconds",
            RASPollMilliseconds, RASPollMilliseconds_fmt);
         GetRegDWORD(k, "RunOnce",
            RunOnce, RunOnce_fmt);
         GetRegDWORD(k, "NTPPollMilliseconds",
            NTPPollMilliseconds, NTPPollMilliseconds_fmt);
         GetRegDWORD(k, "MaxAdjustSeconds",
            MaxAdjustSeconds, MaxAdjustSeconds_fmt);
         use_MaxAdjustSeconds = (MaxAdjustSeconds > 0);
         GetRegDWORD(k, "AllowPoke",
            AllowPoke, AllowPoke_fmt);
         GetRegDWORD(k, "AllowMonitor",
            AllowMonitor, AllowMonitor_fmt);
         GetRegDWORD(k, "RetryMaxCount",
            RetryMaxCount, RetryMaxCount_fmt);
         use_RetryMaxCount = (RetryMaxCount > 0);
         GetRegDWORD(k, "RetryMinMilliseconds",
            RetryMinMilliseconds, RetryMinMilliseconds_fmt);
         GetRegDWORD(k, "RetryMaxMilliseconds",
            RetryMaxMilliseconds, RetryMaxMilliseconds_fmt);
         GetRegDWORD(k, "RetryAdd",
            RetryAdd, RetryAdd_fmt);
         GetRegDWORD(k, "RetryMult",
            RetryMult, RetryMult_fmt);
         GetRegDWORD(k, "RetryDiv",
            RetryDiv, RetryDiv_fmt);
         GetRegDWORD(k, "NtpProtocolVersion",
            NtpProtocolVersion, NtpProtocolVersion_fmt);
         use_NtpProtocolVersion = (NtpProtocolVersion != 3);
         GetRegDWORD(k, "StartupDelayMilliseconds",
            StartupDelayMilliseconds, StartupDelayMilliseconds_fmt);
         GetRegDWORD(k, "RasDelayMilliseconds",
            RasDelayMilliseconds, RasDelayMilliseconds_fmt);
         GetRegDWORD(k, "ShutdownDelayMilliseconds",
            ShutdownDelayMilliseconds, ShutdownDelayMilliseconds_fmt);
         GetRegDWORD(k, "DisablePacketErrorChecks",
            DisablePacketErrorChecks, DisablePacketErrorChecks_fmt);
         GetRegDWORD(k, "NtpReplyPort",
            NtpReplyPort, NtpReplyPort_fmt);
         use_NtpReplyPort = (NtpReplyPort != 0);
         if (! use_NtpReplyPort) NtpReplyPort = 123;
         RegCloseKey(k);
         break;
      }
   }
}

// Use FormatMessage to compose an error message and popup a message box
//
static void
ShowError(HWND hwnd, const char * msg, DWORD last_err) {
   char msgbuf[4000];
   int len = strlen(msg);
   int left = sizeof(msgbuf);
   if (len > left) {
      len = left - (sizeof(msgbuf) / 2);
   }
   memcpy((void *)msgbuf, (void *)msg, len);
   if (msgbuf[len - 1] != '\n') {
      msgbuf[len++] = '\n';
   }
   left -= len;
   DWORD size = FormatMessage(
                   FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL,
                   last_err,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   &msgbuf[len],
                   left,
                   NULL);
   if (size == 0) {
      wsprintf(&msgbuf[len], "Unknown error code: %#08x", last_err);
      size = strlen(&msgbuf[len]);
   }
   len += size;
   if (msgbuf[len - 1] != '\n') {
      msgbuf[len++] = '\n';
   }
   msgbuf[len++] = '\0';
   for (int i = 0; i < len; ++i) {
      if (msgbuf[i] == '\r') {
         msgbuf[i] = ' ';
      }
   }
   MessageBox(hwnd, msgbuf, "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
}

// Check on a directory name, offer to create it and any required
// subdirectories if it doesn't exit. Rrturn true if everything worked,
// false if don't like that directory name.
//
static bool
CheckDirectoryName(HWND hwnd, char * dn) {
   bool rval = true;
   std::vector<std::string> dirs;
   std::string dir;
   char c;
   for ( ; ; ) {
      c = *dn++;
      if ((c == '/') || (c == '\\') || (c == '\0')) {
         if (dir.size() > 0) {
            dir.append(1, '\\');
            dirs.push_back(dir);
            dir.erase();
         }
         if (c == '\0') {
            break;
         }
      } else {
         dir.append(1, c);
      }
   }
   bool ask_to_create = true;
   for (int i = 0; i < dirs.size(); ++i) {
      dir.append(dirs[i]);
      DWORD attr = GetFileAttributes(dir.c_str());
      if (attr == 0xFFFFFFFF) {
         DWORD code = GetLastError();
         if ((code == ERROR_FILE_NOT_FOUND) ||
             (code == ERROR_PATH_NOT_FOUND)) {
            if (ask_to_create) {
               std::string msg;
               msg.append(dir);
               msg.append(" does not exist\nshould I create it");
               if (i < (dirs.size() - 1)) {
                  msg.append(" and subtree ");
                  for (int j = i + 1; j < dirs.size(); ++j) {
                     msg.append(dirs[j]);
                  }
               }
               msg.append(1, '?');
               if (MessageBox(hwnd, msg.c_str(), "Log Directory Error",
                      MB_OKCANCEL | MB_ICONQUESTION) != IDOK) {
                  return false;
               }
               ask_to_create = false;
            }
            if (! CreateDirectory(dir.c_str(), NULL)) {
               DWORD errcode = GetLastError();
               std::string msg("CreateDirectory(");
               msg.append(dir);
               msg.append(")");
               ShowError(hwnd, msg.c_str(), errcode);
               return false;
            }
         } else {
            DWORD errcode = GetLastError();
            std::string msg("GetFileAttributes(");
            msg.append(dir);
            msg.append(")");
            ShowError(hwnd, msg.c_str(), errcode);
            return false;
         }
      } else if ((attr & FILE_ATTRIBUTE_DIRECTORY) == 0) {
         char msg[1024];
         wsprintf(msg, "%s exists, but is not a directory", dir.c_str());
         MessageBox(hwnd, msg, "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
         return false;
      }
   }
   return true;
}

// WriteNTPTimeRegistry is called (naturally) to write the edited values
// back out to the registry (the apply and ok buttons trigger this).
//
static bool
WriteNTPTimeRegistry(HWND hwnd) {

   // If nothing changed, don't bother

   if (! changed) {
      return true;
   }

   // Get local copies of everything and make any required adjustments
   // before writing to the registry

   std::vector<std::string> & servers = GetServerNameList();
   std::string temp;
   int servlen = servers.size();
   for (int i = 0; i < servlen; ++i) {
      temp.append(servers[i]);
      if (i != (servlen - 1)) {
         temp.append("; ");
      }
   }
   const char * my_NTPServerName = temp.c_str();
   if (my_NTPServerName == NULL) my_NTPServerName = "";

   char * my_LogDirectory = LogDirectory;
   if (my_LogDirectory == NULL) my_LogDirectory = "";
   DWORD my_WriteLogEntries = WriteLogEntries;
   DWORD my_LogFileCount = LogFileCount;
   if (! use_LogFileCount) my_LogFileCount = 0;
   DWORD my_NTPServerPort = NTPServerPort;
   if (! use_NTPServerPort) my_NTPServerPort = 123;
   DWORD my_WaitForRAS = WaitForRAS;
   DWORD my_RASPollMilliseconds = RASPollMilliseconds;
   DWORD my_NTPPollMilliseconds = NTPPollMilliseconds;
   DWORD my_RunOnce = RunOnce;
   DWORD my_MaxAdjustSeconds = MaxAdjustSeconds;
   if (! use_MaxAdjustSeconds) my_MaxAdjustSeconds = 0;
   DWORD my_AllowPoke = AllowPoke;
   DWORD my_AllowMonitor = AllowMonitor;
   DWORD my_RetryMaxCount = RetryMaxCount;
   if (! use_RetryMaxCount) my_RetryMaxCount = 0;
   DWORD my_RetryMinMilliseconds = RetryMinMilliseconds;
   DWORD my_RetryMaxMilliseconds = RetryMaxMilliseconds;
   DWORD my_RetryAdd = RetryAdd;
   DWORD my_RetryMult = RetryMult;
   DWORD my_RetryDiv = RetryDiv;
   if (my_RetryDiv <= 0) my_RetryDiv = 1;
   DWORD my_NtpProtocolVersion = NtpProtocolVersion;
   if (! use_NtpProtocolVersion) my_NtpProtocolVersion = 3;
   DWORD my_StartupDelayMilliseconds = StartupDelayMilliseconds;
   DWORD my_RasDelayMilliseconds = RasDelayMilliseconds;
   DWORD my_ShutdownDelayMilliseconds = ShutdownDelayMilliseconds;
   DWORD my_DisablePacketErrorChecks = DisablePacketErrorChecks;
   DWORD my_NtpReplyPort = NtpReplyPort;
   if (! use_NtpReplyPort) my_NtpReplyPort = 0;

   bool err = false;
   if ((my_NTPServerName == NULL) || (*my_NTPServerName == '\0')) {
      MessageBox(hwnd, "You must specify at least one NTP server name.",
                 "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
      err = true;
   }
   if (my_WriteLogEntries) {
      if ((my_LogDirectory == NULL) || (*my_LogDirectory == '\0')) {
         MessageBox(hwnd, "You must specify a directory for log files.",
                    "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
         err = true;
      } else {
         if (! CheckDirectoryName(hwnd, my_LogDirectory)) {
            err = true;
         }
      }
   }
   if (RetryDiv <= 0) {
      MessageBox(hwnd, "Retry divisor must be greater than zero",
         "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
      err = true;
   }
   if (RetryMult <= 0) {
      MessageBox(hwnd, "Retry multiplier must be greater than zero",
         "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
      err = true;
   }
   if (RetryMinMilliseconds > RetryMaxMilliseconds) {
      MessageBox(hwnd, "Min retry time must be less than or equal to max time",
         "NTPTime Error", MB_OK | MB_ICONEXCLAMATION);
      err = true;
   }
   if (err) {
      return false;
   }
   HKEY ntpkey;
   if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,
                      "SOFTWARE\\ntcli\\NTPTime\\4.3",
                      0,
                      "",
                      REG_OPTION_NON_VOLATILE,
                      KEY_ALL_ACCESS,
                      NULL,
                      &ntpkey,
                      NULL) == ERROR_SUCCESS) {

      if (RegSetValueEx(ntpkey, "NTPServerName", 0, REG_SZ,
                        (CONST BYTE*)my_NTPServerName,
                        strlen(my_NTPServerName) + 1) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"NTPServerName\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "LogDirectory", 0, LogDirectory_fmt,
                        (CONST BYTE*)my_LogDirectory,
                        strlen(my_LogDirectory) + 1) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"LogDirectory\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "WriteLogEntries", 0, REG_DWORD,
                        (CONST BYTE*)&my_WriteLogEntries,
                        sizeof(my_WriteLogEntries)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"WriteLogEntries\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "LogFileCount", 0, REG_DWORD,
                        (CONST BYTE*)&my_LogFileCount,
                        sizeof(my_LogFileCount)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"LogFileCount\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "NTPServerPort", 0, REG_DWORD,
                        (CONST BYTE*)&my_NTPServerPort,
                        sizeof(my_NTPServerPort)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"NTPServerPort\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "WaitForRAS", 0, REG_DWORD,
                        (CONST BYTE*)&my_WaitForRAS,
                        sizeof(my_WaitForRAS)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"WaitForRAS\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RASPollMilliseconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_RASPollMilliseconds,
                        sizeof(my_RASPollMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RASPollMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "NTPPollMilliseconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_NTPPollMilliseconds,
                        sizeof(my_NTPPollMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"NTPPollMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RunOnce", 0, REG_DWORD,
                        (CONST BYTE*)&my_RunOnce,
                        sizeof(my_RunOnce)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RunOnce\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "MaxAdjustSeconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_MaxAdjustSeconds,
                        sizeof(my_MaxAdjustSeconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"MaxAdjustSeconds\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "AllowPoke", 0, REG_DWORD,
                        (CONST BYTE*)&my_AllowPoke,
                        sizeof(my_AllowPoke)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"AllowPoke\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "AllowMonitor", 0, REG_DWORD,
                        (CONST BYTE*)&my_AllowMonitor,
                        sizeof(my_AllowMonitor)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"AllowMonitor\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RetryMaxCount", 0, REG_DWORD,
                        (CONST BYTE*)&my_RetryMaxCount,
                        sizeof(my_RetryMaxCount)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RetryMaxCount\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RetryMinMilliseconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_RetryMinMilliseconds,
                        sizeof(my_RetryMinMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RetryMinMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RetryMaxMilliseconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_RetryMaxMilliseconds,
                        sizeof(my_RetryMaxMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RetryMaxMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RetryAdd", 0, REG_DWORD,
                        (CONST BYTE*)&my_RetryAdd,
                        sizeof(my_RetryAdd)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RetryAdd\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RetryMult", 0, REG_DWORD,
                        (CONST BYTE*)&my_RetryMult,
                        sizeof(my_RetryMult)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RetryMult\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RetryDiv", 0, REG_DWORD,
                        (CONST BYTE*)&my_RetryDiv,
                        sizeof(my_RetryDiv)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RetryDiv\")", GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "NtpProtocolVersion", 0, REG_DWORD,
                        (CONST BYTE*)&my_NtpProtocolVersion,
                        sizeof(my_NtpProtocolVersion)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"NtpProtocolVersion\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "StartupDelayMilliseconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_StartupDelayMilliseconds,
                        sizeof(my_StartupDelayMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"StartupDelayMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "RasDelayMilliseconds", 0, REG_DWORD,
                        (CONST BYTE*)&my_RasDelayMilliseconds,
                        sizeof(my_RasDelayMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"RasDelayMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "ShutdownDelayMilliseconds", 0, REG_DWORD,
                     (CONST BYTE*)&my_ShutdownDelayMilliseconds,
                     sizeof(my_ShutdownDelayMilliseconds)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"ShutdownDelayMilliseconds\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "DisablePacketErrorChecks", 0, REG_DWORD,
                     (CONST BYTE*)&my_DisablePacketErrorChecks,
                     sizeof(my_DisablePacketErrorChecks)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"DisablePacketErrorChecks\")",
                   GetLastError());
         err = true;
      }
      if (RegSetValueEx(ntpkey, "NtpReplyPort", 0, REG_DWORD,
                     (CONST BYTE*)&my_NtpReplyPort,
                     sizeof(my_NtpReplyPort)) != ERROR_SUCCESS) {
         ShowError(hwnd, "RegSetValueEx(\"NtpReplyPort\")",
                   GetLastError());
         err = true;
      }
      RegCloseKey(ntpkey);
      if (! err) changed = false;
   } else {
      ShowError(hwnd,
"RegCreateKeyEx(\"HKEY_LOCAL_MACHINE\\SOFTWARE\\ntcli\\NTPTime\\4.3\")",
         GetLastError());
      err = true;
   }
   return ! err;
}

// SetComboMsThruHour - given a combo box, set the possible time units
// in it starting at milliseconds.
//
static void
SetComboMsThruHour(HWND hDlg, int cmbo_item) {
   HWND cmbo = GetDlgItem(hDlg, cmbo_item);
   SendMessage(cmbo, CB_RESETCONTENT, 0, 0);
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Milliseconds");
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Seconds");
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Minutes");
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Hours");
   SendMessage(cmbo, CB_SETCURSEL, 0, 0);
}

// SetComboSecThruHour - similar to SetComboMsThruHour, but leaves out
// the milliseconds choice and starts at seconds.
//
static void
SetComboSecThruHour(HWND hDlg, int cmbo_item) {
   HWND cmbo = GetDlgItem(hDlg, cmbo_item);
   SendMessage(cmbo, CB_RESETCONTENT, 0, 0);
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Seconds");
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Minutes");
   SendMessage(cmbo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) "Hours");
   SendMessage(cmbo, CB_SETCURSEL, 0, 0);
}

// SetSpinner - initialize a spinner so it will start at 0 and go up.
//
static void
SetSpinner(HWND hDlg, int spin_item, int min_val = 0) {
   SendMessage(GetDlgItem(hDlg, spin_item),
      UDM_SETRANGE, 0, (LPARAM) MAKELONG((short) UD_MAXVAL, (short) min_val));
}

// SetMsValAndUnits - given a spinner and a combo box which allows
// milliseconds thru hours as the units, convert a millisecond value to the
// most appropriate units for display and set the spinner and combo box to
// reflect that choice.
//
static void
SetMsValAndUnits(HWND hDlg, int spinner, int combo, DWORD ms) {
   int units = 0;
   if ((ms > 0) && ((ms % 1000) == 0)) {
      units = 1;
      ms /= 1000;
      if ((ms % 60) == 0) {
         units = 2;
         ms /= 60;
         if ((ms % 60) == 0) {
            units = 3;
            ms /= 60;
         }
      }
   }
   SendMessage(GetDlgItem(hDlg, spinner), UDM_SETPOS,
      0, (LPARAM)MAKELONG((short)ms, 0));
   SendMessage(GetDlgItem(hDlg, combo), CB_SETCURSEL, units, 0);
}

// SetSecValAndUnits - similar to SetMsValAndUnits, but works on spinner and
// combo pair that start at seconds rather than milliseconds.
//
static void
SetSecValAndUnits(HWND hDlg, int spinner, int combo, DWORD sec) {
   int units = 0;
   if ((sec > 0) && ((sec % 60) == 0)) {
      units = 1;
      sec /= 60;
      if ((sec % 60) == 0) {
         units = 2;
         sec /= 60;
      }
   }
   SendMessage(GetDlgItem(hDlg, spinner), UDM_SETPOS,
      0, (LPARAM)MAKELONG((short)sec, 0));
   SendMessage(GetDlgItem(hDlg, combo), CB_SETCURSEL, units, 0);
}

// EnableMaxAdjust - based on the value of MaxAdjustSeconds, enable max time
// windows appropriately.
//
static void
EnableMaxAdjust(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_STATIC_SANITY), use_MaxAdjustSeconds);
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_SANITY), use_MaxAdjustSeconds);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_SANITY), use_MaxAdjustSeconds);
   EnableWindow(GetDlgItem(hDlg, IDC_COMBO_SANITY), use_MaxAdjustSeconds);
}

// EnableRasPoll - based on the WaitForRAS flag, enable ras polling windows
// appropriately.
//
static void
EnableRasPoll(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_STATIC_RASWAIT), WaitForRAS);
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_RASWAIT), WaitForRAS);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_RASWAIT), WaitForRAS);
   EnableWindow(GetDlgItem(hDlg, IDC_COMBO_RASWAIT), WaitForRAS);
}

// EnableRunOnce - based on the RunOnce flag, enable ntp polling windows
// appropriately.
//
static void
EnableRunOnce(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_POLL), ! RunOnce);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_POLL), ! RunOnce);
   EnableWindow(GetDlgItem(hDlg, IDC_COMBO_POLL), ! RunOnce);
}

static bool
FetchNumberField(HWND hDlg, int item, DWORD & num) {
   int len = GetWindowTextLength(GetDlgItem(hDlg, item));
   if (len > 0) {
      char * txt = new char [len + 1];
      GetWindowText(GetDlgItem(hDlg, item), txt, len + 1);
      num = atoi(txt);
      delete [] txt;
      return true;
   }
   return false;
}

static void
FetchMSNumberField(HWND hDlg, int txt_item, int combo_item, DWORD & num) {
   if (FetchNumberField(hDlg, txt_item, num)) {
      int units = (int)SendMessage(GetDlgItem(hDlg, combo_item),
                                      CB_GETCURSEL, 0, 0);
      switch(units) {
      case 1:
         num = num * 1000;
         break;
      case 2:
         num = num * 1000 * 60;
         break;
      case 3:
         num = num * 1000 * 60 * 60;
         break;
      }
   }
}

static void
FetchSecNumberField(HWND hDlg, int txt_item, int combo_item, DWORD & num) {
   if (FetchNumberField(hDlg, txt_item, num)) {
      int units = (int)SendMessage(GetDlgItem(hDlg, combo_item),
                                      CB_GETCURSEL, 0, 0);
      switch(units) {
      case 1:
         num = num * 60;
         break;
      case 2:
         num = num * 60 * 60;
         break;
      }
   }
}

static void
EnableServerButtons(HWND hDlg) {
   bool enable_delete, enable_up, enable_down;
   int i = SendMessage(GetDlgItem(hDlg, IDC_LIST_SERVERS), LB_GETCURSEL, 0, 0); 
   if (i == LB_ERR) {
      selected_server = -1;
      enable_delete = false;
      enable_up = false;
      enable_down = false;
   } else {
      selected_server = i;
      enable_delete = true;
      enable_up = (i > 0);
      enable_down = (i < (GetServerNameList().size() - 1));
   }
   EnableWindow(GetDlgItem(hDlg, IDC_BTN_UP_SERVERS), enable_up);
   EnableWindow(GetDlgItem(hDlg, IDC_BTN_DOWN_SERVERS), enable_down);
   EnableWindow(GetDlgItem(hDlg, IDC_BTN_DEL_SERVERS), enable_delete);
   EnableWindow(GetDlgItem(hDlg, IDC_BTN_ADD_SERVERS),
      (GetWindowTextLength(GetDlgItem(hDlg, IDC_EDIT_SERVERS)) > 0));
}

static void
DrawServersList(HWND hDlg) {
   HWND list_ctl = GetDlgItem(hDlg, IDC_LIST_SERVERS);
   SendMessage(list_ctl, LB_RESETCONTENT, 0, 0);
   std::vector<std::string> & servers = GetServerNameList();
   int len = servers.size();
   for (int i = 0; i < len; ++i) {
      const char * str = servers[i].c_str();
      SendMessage(list_ctl, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str); 
   }
   SendMessage(list_ctl, LB_SETCURSEL, (WPARAM)selected_server, 0);
}

// ServersProc - the dialog procedure for the Servers page. Initializes
// the list of servers from the server name string and process all the
// controls to edit the string, reorder entries, etc.
//
static BOOL APIENTRY
ServersProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
   switch (message) {

   case WM_INITDIALOG:
      {
         if ((NTPServerName != NULL) && (*NTPServerName != '\0')) {
            AppendServerNames(NTPServerName);
         }
         DrawServersList(hDlg);
         EnableServerButtons(hDlg);
      }
      return TRUE;

   case WM_COMMAND:
      switch (LOWORD(wParam)) {
      case IDC_LIST_SERVERS:
         if (HIWORD(wParam) == LBN_SELCHANGE) {
            EnableServerButtons(hDlg);
            return TRUE;
         }
         break;
      case IDC_BTN_UP_SERVERS: {
         std::vector<std::string> & servers = GetServerNameList();
         std::string t1(servers[selected_server - 1]);
         std::string t2(servers[selected_server]);
         servers[selected_server - 1] = t2;
         servers[selected_server] = t1;
         --selected_server;
         DrawServersList(hDlg);
         EnableServerButtons(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         } return TRUE;
      case IDC_BTN_DOWN_SERVERS: {
         std::vector<std::string> & servers = GetServerNameList();
         std::string t1(servers[selected_server + 1]);
         std::string t2(servers[selected_server]);
         servers[selected_server + 1] = t2;
         servers[selected_server] = t1;
         ++selected_server;
         DrawServersList(hDlg);
         EnableServerButtons(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         } return TRUE;
      case IDC_BTN_DEL_SERVERS: {
         std::vector<std::string> & servers = GetServerNameList();
         servers.erase(servers.begin() + selected_server);
         selected_server = -1;
         DrawServersList(hDlg);
         EnableServerButtons(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         } return TRUE;
      case IDC_EDIT_SERVERS:
         EnableServerButtons(hDlg);
         return TRUE;
      case IDC_BTN_ADD_SERVERS:
         int len = GetWindowTextLength(GetDlgItem(hDlg, IDC_EDIT_SERVERS));
         if (len > 0) {
            char * newnames = new char [len + 1];
            GetWindowText(GetDlgItem(hDlg, IDC_EDIT_SERVERS),
               newnames, len + 1);
            AppendServerNames(newnames);
            selected_server = GetServerNameList().size() - 1;
            delete [] newnames;
            SetWindowText(GetDlgItem(hDlg, IDC_EDIT_SERVERS), "");
            DrawServersList(hDlg);
            EnableServerButtons(hDlg);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         return TRUE;
      }
      break;

   case WM_NOTIFY: {
      LPNMHDR pnmh = (LPNMHDR)lParam;
      switch (pnmh->code) {
      case PSN_APPLY:
         if (! WriteNTPTimeRegistry(hDlg)) {
            SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE);
         }
         break;
      }
      } return TRUE;
   }
   return FALSE;
}

// SettingsProc - the dialog procedure for the settings page. Initializes
// everything from the registry values, then modifies the globals as the
// user makes changes.
//
static BOOL APIENTRY
SettingsProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
   switch (message) {
   case WM_INITDIALOG:

      // Initialize units in each of the combo boxes.

      SetComboSecThruHour(hDlg, IDC_COMBO_SANITY);
      SetComboMsThruHour(hDlg, IDC_COMBO_RASWAIT);
      SetComboMsThruHour(hDlg, IDC_COMBO_POLL);

      // Setup all the spinners to run from zero to infinity

      SetSpinner(hDlg, IDC_SPIN_SANITY);
      SetSpinner(hDlg, IDC_SPIN_RASWAIT);
      SetSpinner(hDlg, IDC_SPIN_POLL);

      // Set all the initial values according to the registry entries

      SetSecValAndUnits(hDlg,
         IDC_SPIN_SANITY, IDC_COMBO_SANITY, MaxAdjustSeconds);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_RASWAIT, IDC_COMBO_RASWAIT, RASPollMilliseconds);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_POLL, IDC_COMBO_POLL, NTPPollMilliseconds);
      SendMessage(GetDlgItem(hDlg, IDC_CHECK_SANITY), BM_SETCHECK,
         (WPARAM)(use_MaxAdjustSeconds), 0);
      SendMessage(GetDlgItem(hDlg, IDC_CHECK_RASWAIT), BM_SETCHECK,
         (WPARAM)(WaitForRAS), 0);
      SendMessage(GetDlgItem(hDlg, IDC_RADIO_SETONCE), BM_SETCHECK,
         (WPARAM)(RunOnce), 0);
      SendMessage(GetDlgItem(hDlg, IDC_RADIO_POLL), BM_SETCHECK,
         (WPARAM)(! RunOnce), 0);
      SendMessage(GetDlgItem(hDlg, IDC_CHECK_POKE), BM_SETCHECK,
         (WPARAM)(AllowPoke), 0);
      SendMessage(GetDlgItem(hDlg, IDC_CHECK_MON), BM_SETCHECK,
         (WPARAM)(AllowMonitor), 0);

      // Enable the various windows according to the current set of choices

      EnableMaxAdjust(hDlg);
      EnableRasPoll(hDlg);
      EnableRunOnce(hDlg);
      return TRUE;

   case WM_COMMAND:
      switch (LOWORD(wParam)) {
      case IDC_CHECK_SANITY: // checkbox for sanity check
         use_MaxAdjustSeconds = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_SANITY), BM_GETCHECK, 0, 0));
         EnableMaxAdjust(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_EDIT_SANITY: // text field for max time
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchSecNumberField(hDlg, IDC_EDIT_SANITY, IDC_COMBO_SANITY,
               MaxAdjustSeconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_SANITY: // combo box for units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchSecNumberField(hDlg, IDC_EDIT_SANITY, IDC_COMBO_SANITY,
               MaxAdjustSeconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_CHECK_RASWAIT: // checkbox for ras waiting
         WaitForRAS =
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_RASWAIT), BM_GETCHECK, 0, 0);
         EnableRasPoll(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_EDIT_RASWAIT: // text box for ras poll time
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_RASWAIT, IDC_COMBO_RASWAIT,
               RASPollMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_RASWAIT: // combo box for ras poll units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_RASWAIT, IDC_COMBO_RASWAIT,
               RASPollMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_RADIO_SETONCE: // radio button for set once
      case IDC_RADIO_POLL: // radio button for set many
         RunOnce =
            SendMessage(GetDlgItem(hDlg, IDC_RADIO_SETONCE), BM_GETCHECK, 0, 0);
         EnableRunOnce(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_EDIT_POLL: // text area for time interval
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_POLL, IDC_COMBO_POLL,
               NTPPollMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_POLL: // combo box for time interval
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_POLL, IDC_COMBO_POLL,
               NTPPollMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_CHECK_POKE: // checkbox for "allow NTPTray set time"
         AllowPoke =
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_POKE), BM_GETCHECK, 0, 0);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_CHECK_MON: // checkbox for "allow NTPTray monitor state"
         AllowMonitor =
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_MON), BM_GETCHECK, 0, 0);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      }
      return TRUE;

   case WM_NOTIFY: {
      LPNMHDR pnmh = (LPNMHDR)lParam;
      switch (pnmh->code) {
      case PSN_APPLY:
         // If the message handler is in a dialog box procedure, you must
         // use the SetWindowLong function with DWL_MSGRESULT to set a
         // return value.
         //
         // LONG SetWindowLong(
         //   HWND hWnd,       // handle to window
         //   int nIndex,      // offset of value to set
         //   LONG dwNewLong   // new value
         // );
         //
         // Return PSNRET_INVALID_NOCHANGEPAGE to prevent the changes from
         // taking effect and to return the focus to the page. Return
         // PSNRET_NOERROR to accept the changes and allow the property
         // sheet to be destroyed. To set the return value, the dialog box
         // procedure for the page must use the SetWindowLong function with
         // the DWL_MSGRESULT value, and the dialog box procedure must
         // return TRUE.

         if (! WriteNTPTimeRegistry(hDlg)) {
            SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE);
         }
         break;
      }
      } return TRUE;
   }
   return FALSE;
}

// EnableRemoveLogWindows - enable the controls in the remove log file area
// based on the status of LogFileCount.
//
static void
EnableRemoveLogWindows(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_REMOVE_LOG),
      WriteLogEntries && use_LogFileCount);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_REMOVE_LOG),
      WriteLogEntries && use_LogFileCount);
}

// EnableLogWindows - enable all the log controls based on WriteLogEntries
//
static void
EnableLogWindows(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_STATIC_LOGDIR), WriteLogEntries);
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_LOGDIR), WriteLogEntries);
   EnableWindow(GetDlgItem(hDlg, IDC_BUTTON_LOGDIR), WriteLogEntries);
   EnableWindow(GetDlgItem(hDlg, IDC_CHECK_REMOVE_LOG), WriteLogEntries);
   EnableRemoveLogWindows(hDlg);
}

// LogFilesProc - the dialog procedure for the log files page
//
static BOOL APIENTRY
LogFilesProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
   switch (message) {

   case WM_INITDIALOG: {
      SendMessage(GetDlgItem(hDlg, IDC_CHECK_ENABLE_LOG), BM_SETCHECK,
         (WPARAM)(WriteLogEntries), 0);
      if (LogDirectory != NULL) {
         SetWindowText(GetDlgItem(hDlg, IDC_EDIT_LOGDIR), LogDirectory);
      }
      SetWindowText(GetDlgItem(hDlg, IDC_EDIT_REMOVE_LOG), "10");
      SendMessage(GetDlgItem(hDlg, IDC_CHECK_REMOVE_LOG), BM_SETCHECK,
         (WPARAM)use_LogFileCount, 0);
      SendMessage(GetDlgItem(hDlg, IDC_SPIN_REMOVE_LOG), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)LogFileCount, 0));
      SetSpinner(hDlg, IDC_SPIN_REMOVE_LOG, 2);
      EnableLogWindows(hDlg);
      } return TRUE;

   case WM_COMMAND:
      switch (LOWORD(wParam)) {
      case IDC_CHECK_ENABLE_LOG:
         WriteLogEntries = (DWORD)
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_ENABLE_LOG), BM_GETCHECK,
                        0, 0);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         EnableLogWindows(hDlg);
         break;
      case IDC_BUTTON_LOGDIR: {
         LPMALLOC lpMalloc;
         SHGetMalloc(&lpMalloc);
         BROWSEINFO bi;
         char da_name[MAX_PATH + 10];
         LPITEMIDLIST pidl;
         memset((void *)&bi, 0, sizeof(bi));
         bi.hwndOwner = hDlg;
         bi.pszDisplayName = da_name;
         bi.lpszTitle = "Select directory for NTPTime log files";
         bi.ulFlags |= BIF_RETURNONLYFSDIRS;
         if ((pidl = SHBrowseForFolder(&bi)) != NULL) {
            if (SHGetPathFromIDList(pidl, da_name)) {
               SetWindowText(GetDlgItem(hDlg, IDC_EDIT_LOGDIR), da_name);
               delete [] LogDirectory;
               LogDirectory = new char [strlen(da_name) + 1];
               strcpy(LogDirectory, da_name);
               LogDirectory_fmt = REG_SZ;
               PropSheet_Changed(GetParent(hDlg), hDlg);
               changed = true;
            }
            lpMalloc->Free(pidl);
         }
         lpMalloc->Release();
         } break;
      case IDC_EDIT_LOGDIR:
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            delete [] LogDirectory;
            LogDirectory = NULL;
            int len = GetWindowTextLength(GetDlgItem(hDlg, IDC_EDIT_LOGDIR));
            if (len > 0) {
               LogDirectory = new char [len + 1];
               GetWindowText(GetDlgItem(hDlg, IDC_EDIT_LOGDIR),
                  LogDirectory, len + 1);
               PropSheet_Changed(GetParent(hDlg), hDlg);
               changed = true;
            }
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            LogDirectory_fmt = REG_SZ;
            changed = true;
         }
         break;
      case IDC_CHECK_REMOVE_LOG: {
         use_LogFileCount = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_REMOVE_LOG), BM_GETCHECK,
                        0, 0));
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         EnableRemoveLogWindows(hDlg);
         } break;
      case IDC_EDIT_REMOVE_LOG:
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_EDIT_REMOVE_LOG, LogFileCount);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      }
      return TRUE;

   case WM_NOTIFY: {
      LPNMHDR pnmh = (LPNMHDR)lParam;
      switch (pnmh->code) {
      case PSN_APPLY:
         if (! WriteNTPTimeRegistry(hDlg)) {
            SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE);
         }
         break;
      }
      } return TRUE;
   }
   return FALSE;
}

static void
EnableMaxError(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_STATIC_MAXERR), use_RetryMaxCount);
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_MAXERR), use_RetryMaxCount);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_MAXERR), use_RetryMaxCount);
}

// RetryProc runs the "Retry" property page.
//
static BOOL APIENTRY
RetryProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
   switch (message) {

   case WM_INITDIALOG: {
      SetComboMsThruHour(hDlg, IDC_COMBO_MIN_RETRY);
      SetComboMsThruHour(hDlg, IDC_COMBO_MAX_RETRY);
      SetComboMsThruHour(hDlg, IDC_COMBO_ADD_RETRY);

      SetSpinner(hDlg, IDC_SPIN_MAXERR);
      SetSpinner(hDlg, IDC_SPIN_MIN_RETRY);
      SetSpinner(hDlg, IDC_SPIN_MAX_RETRY);
      SetSpinner(hDlg, IDC_SPIN_ADD_RETRY);
      SetSpinner(hDlg, IDC_SPIN_MUL_RETRY, 1);
      SetSpinner(hDlg, IDC_SPIN_DIV_RETRY, 1);

      SendMessage(GetDlgItem(hDlg, IDC_CHECK_MAXERR), BM_SETCHECK,
         (WPARAM)(use_RetryMaxCount), 0);
      SendMessage(GetDlgItem(hDlg, IDC_SPIN_MAXERR), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)RetryMaxCount, 0));
      EnableMaxError(hDlg);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_MIN_RETRY, IDC_COMBO_MIN_RETRY, RetryMinMilliseconds);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_MAX_RETRY, IDC_COMBO_MAX_RETRY, RetryMaxMilliseconds);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_ADD_RETRY, IDC_COMBO_ADD_RETRY, RetryAdd);
      SendMessage(GetDlgItem(hDlg, IDC_SPIN_MUL_RETRY), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)RetryMult, 0));
      SendMessage(GetDlgItem(hDlg, IDC_SPIN_DIV_RETRY), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)RetryDiv, 0));
      } return TRUE;

   case WM_COMMAND:
      switch (LOWORD(wParam)) {
      case IDC_CHECK_MAXERR: // checkbox for max error count
         use_RetryMaxCount = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_MAXERR), BM_GETCHECK, 0, 0));
         EnableMaxError(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_EDIT_MAXERR: // text box for max error count
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_EDIT_MAXERR, RetryMaxCount);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_MIN_RETRY: // text box for min retry time
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_MIN_RETRY, IDC_COMBO_MIN_RETRY,
               RetryMinMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_MIN_RETRY: // combo box for min retry time units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_MIN_RETRY, IDC_COMBO_MIN_RETRY,
               RetryMinMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_MAX_RETRY: // text box for max retry time
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_MAX_RETRY, IDC_COMBO_MAX_RETRY,
               RetryMaxMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_MAX_RETRY: // combo box for max retry time units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_MAX_RETRY, IDC_COMBO_MAX_RETRY,
               RetryMaxMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_ADD_RETRY: // text box for retry add time
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_ADD_RETRY, IDC_COMBO_ADD_RETRY,
               RetryAdd);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_ADD_RETRY: // combo box for add retry time units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_ADD_RETRY, IDC_COMBO_ADD_RETRY,
               RetryAdd);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_MUL_RETRY: // text box for retry multiplier
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_EDIT_MUL_RETRY, RetryMult);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_DIV_RETRY: // text box for retry divider
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_EDIT_DIV_RETRY, RetryDiv);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      }
      return TRUE;

   case WM_NOTIFY: {
      LPNMHDR pnmh = (LPNMHDR)lParam;
      switch (pnmh->code) {
      case PSN_APPLY:
         if (! WriteNTPTimeRegistry(hDlg)) {
            SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE);
         }
         break;
      }
      } return TRUE;
   }
   return FALSE;
}

static void
EnableNtpProtocol(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_STATIC_NTP_PROTOCOL),
                use_NtpProtocolVersion);
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_NTP_PROTOCOL),
                use_NtpProtocolVersion);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_NTP_PROTOCOL),
                use_NtpProtocolVersion);
}

static void
EnableNtpPort(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_STATIC_NTP_PORT), use_NTPServerPort);
   EnableWindow(GetDlgItem(hDlg, IDC_EDIT_NTP_PORT), use_NTPServerPort);
   EnableWindow(GetDlgItem(hDlg, IDC_SPIN_NTP_PORT), use_NTPServerPort);
}

// AdvancedProc runs the "Advanced" property page.
//
static BOOL APIENTRY
AdvancedProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
   switch (message) {

   case WM_INITDIALOG: {
      SetComboMsThruHour(hDlg, IDC_COMBO_STARTUP_DELAY);
      SetComboMsThruHour(hDlg, IDC_COMBO_DIALUP_DELAY);
      SetComboMsThruHour(hDlg, IDC_COMBO_SET_ONCE_DELAY);

      SetSpinner(hDlg, IDC_SPIN_STARTUP_DELAY);
      SetSpinner(hDlg, IDC_SPIN_DIALUP_DELAY);
      SetSpinner(hDlg, IDC_SPIN_SET_ONCE_DELAY);
      SetSpinner(hDlg, IDC_SPIN_NTP_PROTOCOL);
      SetSpinner(hDlg, IDC_SPIN_NTP_PORT);

      SetMsValAndUnits(hDlg,
         IDC_SPIN_STARTUP_DELAY, IDC_COMBO_STARTUP_DELAY,
         StartupDelayMilliseconds);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_DIALUP_DELAY, IDC_COMBO_DIALUP_DELAY, RasDelayMilliseconds);
      SetMsValAndUnits(hDlg,
         IDC_SPIN_SET_ONCE_DELAY, IDC_COMBO_SET_ONCE_DELAY,
         ShutdownDelayMilliseconds);

      SendMessage(GetDlgItem(hDlg, IDC_CHECK_NTP_PROTOCOL), BM_SETCHECK,
         (WPARAM)(use_NtpProtocolVersion), 0);
      SendMessage(GetDlgItem(hDlg, IDC_SPIN_NTP_PROTOCOL), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)NtpProtocolVersion, 0));
      EnableNtpProtocol(hDlg);

      SendMessage(GetDlgItem(hDlg, IDC_CHECK_NTP_PORT), BM_SETCHECK,
         (WPARAM)(use_NTPServerPort), 0);
      SendMessage(GetDlgItem(hDlg, IDC_SPIN_NTP_PORT), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)NTPServerPort, 0));
      EnableNtpPort(hDlg);
      } return TRUE;

   case WM_COMMAND:
      switch (LOWORD(wParam)) {
      case IDC_EDIT_STARTUP_DELAY: // text box for startup delay
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_STARTUP_DELAY,
               IDC_COMBO_STARTUP_DELAY, StartupDelayMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_STARTUP_DELAY: // combo box for startup delay units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_STARTUP_DELAY,
               IDC_COMBO_STARTUP_DELAY, StartupDelayMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_DIALUP_DELAY: // text box for dialup delay
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_DIALUP_DELAY,
               IDC_COMBO_DIALUP_DELAY, RasDelayMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_DIALUP_DELAY: // combo box for dialup delay units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_DIALUP_DELAY,
               IDC_COMBO_DIALUP_DELAY, RasDelayMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_EDIT_SET_ONCE_DELAY: // text box for set once delay
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchMSNumberField(hDlg, IDC_EDIT_SET_ONCE_DELAY,
               IDC_COMBO_SET_ONCE_DELAY, ShutdownDelayMilliseconds);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_COMBO_SET_ONCE_DELAY: // text box for set once delay units
         if (HIWORD(wParam) == CBN_SELCHANGE) {
            FetchMSNumberField(hDlg, IDC_EDIT_SET_ONCE_DELAY,
               IDC_COMBO_SET_ONCE_DELAY, ShutdownDelayMilliseconds);
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_CHECK_NTP_PROTOCOL: // check box for using weird protocol version
         use_NtpProtocolVersion = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_NTP_PROTOCOL), BM_GETCHECK,
                        0, 0));
         EnableNtpProtocol(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_EDIT_NTP_PROTOCOL: // text box for protocol number
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_EDIT_NTP_PROTOCOL, NtpProtocolVersion);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      case IDC_CHECK_NTP_PORT: // check box for using weird port number
         use_NTPServerPort = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_CHECK_NTP_PORT), BM_GETCHECK,
                        0, 0));
         EnableNtpPort(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_EDIT_NTP_PORT: // text box for port number
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_EDIT_NTP_PORT, NTPServerPort);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      }
      return TRUE;

   case WM_NOTIFY: {
      LPNMHDR pnmh = (LPNMHDR)lParam;
      switch (pnmh->code) {
      case PSN_APPLY:
         if (! WriteNTPTimeRegistry(hDlg)) {
            SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE);
         }
         break;
      }
      } return TRUE;
   }
   return FALSE;
}

static void
EnableReplyPort(HWND hDlg) {
   EnableWindow(GetDlgItem(hDlg, IDC_REPLY_PORT_TEXT), use_NtpReplyPort);
   EnableWindow(GetDlgItem(hDlg, IDC_REPLY_PORT_NUM), use_NtpReplyPort);
   EnableWindow(GetDlgItem(hDlg, IDC_REPLY_PORT_SPIN), use_NtpReplyPort);
}

// ObscureProc runs the "Obscure" property page.
//
static BOOL APIENTRY
ObscureProc(HWND hDlg, UINT message, UINT wParam, LONG lParam) {
   switch (message) {

   case WM_INITDIALOG: {
      SetSpinner(hDlg, IDC_REPLY_PORT_SPIN);
      SendMessage(GetDlgItem(hDlg, IDC_DETAILED_ERROR_CHECKS), BM_SETCHECK,
         (WPARAM)(DisablePacketErrorChecks), 0);
      SendMessage(GetDlgItem(hDlg, IDC_REPLY_PORT), BM_SETCHECK,
         (WPARAM)(use_NtpReplyPort), 0);
      SendMessage(GetDlgItem(hDlg, IDC_REPLY_PORT_SPIN), UDM_SETPOS,
         0, (LPARAM)MAKELONG((short)NtpReplyPort, 0));
      EnableReplyPort(hDlg);
      } return TRUE;

   case WM_COMMAND:
      switch (LOWORD(wParam)) {
      case IDC_DETAILED_ERROR_CHECKS:
         DisablePacketErrorChecks = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_DETAILED_ERROR_CHECKS),
                        BM_GETCHECK, 0, 0));
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_REPLY_PORT:
         use_NtpReplyPort = (0 !=
            SendMessage(GetDlgItem(hDlg, IDC_REPLY_PORT), BM_GETCHECK,
                        0, 0));
         EnableReplyPort(hDlg);
         PropSheet_Changed(GetParent(hDlg), hDlg);
         changed = true;
         break;
      case IDC_REPLY_PORT_NUM: // text box for port number
         if (HIWORD(wParam) == EN_KILLFOCUS) {
            FetchNumberField(hDlg, IDC_REPLY_PORT_NUM, NtpReplyPort);
         } else if (HIWORD(wParam) == EN_CHANGE) {
            PropSheet_Changed(GetParent(hDlg), hDlg);
            changed = true;
         }
         break;
      }
      return TRUE;

   case WM_NOTIFY: {
      LPNMHDR pnmh = (LPNMHDR)lParam;
      switch (pnmh->code) {
      case PSN_APPLY:
         if (! WriteNTPTimeRegistry(hDlg)) {
            SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE);
         }
         break;
      }
      } return TRUE;
   }
   return FALSE;
}

#define NUM_PAGES 6

static int
RunNTPTimeControlPanel(HINSTANCE hinst, HWND hwndOwner)
{
   ReadNTPTimeRegistry();

   PROPSHEETPAGE psp[NUM_PAGES];
   PROPSHEETHEADER psh;

   int pn = 0;

   memset((void *)&psh, 0, sizeof(psh));
   memset((void *)&psp[0], 0, sizeof(psp[0]) * NUM_PAGES);

   psp[pn].dwSize = sizeof(PROPSHEETPAGE);
   psp[pn].dwFlags = PSP_USETITLE;
   psp[pn].hInstance = hinst;
   psp[pn].pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_SERVERS);
   psp[pn].pszIcon = NULL;
   psp[pn].pfnDlgProc = ServersProc;
   psp[pn].pszTitle = "Servers";
   psp[pn].lParam = 0;

   ++pn;
   psp[pn].dwSize = sizeof(PROPSHEETPAGE);
   psp[pn].dwFlags = PSP_USETITLE;
   psp[pn].hInstance = hinst;
   psp[pn].pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_OPS);
   psp[pn].pszIcon = NULL;
   psp[pn].pfnDlgProc = SettingsProc;
   psp[pn].pszTitle = "Settings";
   psp[pn].lParam = 0;

   ++pn;
   psp[pn].dwSize = sizeof(PROPSHEETPAGE);
   psp[pn].dwFlags = PSP_USETITLE;
   psp[pn].hInstance = hinst;
   psp[pn].pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_LOG);
   psp[pn].pszIcon = NULL;
   psp[pn].pfnDlgProc = LogFilesProc;
   psp[pn].pszTitle = "Log Files";
   psp[pn].lParam = 0;

   ++pn;
   psp[pn].dwSize = sizeof(PROPSHEETPAGE);
   psp[pn].dwFlags = PSP_USETITLE;
   psp[pn].hInstance = hinst;
   psp[pn].pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_RETRY);
   psp[pn].pszIcon = NULL;
   psp[pn].pfnDlgProc = RetryProc;
   psp[pn].pszTitle = "Retry";
   psp[pn].lParam = 0;

   ++pn;
   psp[pn].dwSize = sizeof(PROPSHEETPAGE);
   psp[pn].dwFlags = PSP_USETITLE;
   psp[pn].hInstance = hinst;
   psp[pn].pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_ADVANCED);
   psp[pn].pszIcon = NULL;
   psp[pn].pfnDlgProc = AdvancedProc;
   psp[pn].pszTitle = "Advanced";
   psp[pn].lParam = 0;

   ++pn;
   psp[pn].dwSize = sizeof(PROPSHEETPAGE);
   psp[pn].dwFlags = PSP_USETITLE;
   psp[pn].hInstance = hinst;
   psp[pn].pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_OBSCURE);
   psp[pn].pszIcon = NULL;
   psp[pn].pfnDlgProc = ObscureProc;
   psp[pn].pszTitle = "Obscure";
   psp[pn].lParam = 0;

   psh.dwSize = sizeof(PROPSHEETHEADER);
   psh.dwFlags = PSH_PROPSHEETPAGE;
   psh.hwndParent = hwndOwner;
   psh.hInstance = hinst;
   psh.pszIcon = NULL;
   psh.pszCaption = (LPSTR)"NTPTime 4.3 Control Panel";
   psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
   psh.ppsp = (LPCPROPSHEETPAGE) &psp;

   int rval = PropertySheet(&psh);

   return rval;
}

// This entry point is invoked by the control panel application when loading
// this particular control panel applet. This is where everything starts.
//
// IMPORTANT NOTE: The exported name of the entry point *must* be just
// exactly "CPlApplet", not any of the various mangled versions the
// compilers produce. This is achieved by the miracle of the ntptime-cpl.def
// file included when linking.
//
extern "C" __declspec( dllexport ) LONG APIENTRY
CPlApplet(HWND hwndCPL, UINT uMsg, LONG lParam1, LONG lParam2) {
   switch (uMsg) {

   case CPL_INIT: {
      // This is called once when the control panel is first loaded, get
      // my module handle so I can pass it to DailogBox later.
      InitCommonControls();
      hinst = GetModuleHandle("ntptime.cpl");
      } return 1;

   case CPL_GETCOUNT:
      // We only have 1 applet to process here so just say that.
      return 1;

   case CPL_INQUIRE: {
      // Note that lParam1 better be zero here, but we don't check
      // because there isn't anything I could do about it if the
      // caller screws up anyway...
      LPCPLINFO ci = (LPCPLINFO)lParam2;
      ci->idIcon = IDI_ICON_NTPCPL;
      ci->idName = IDS_ICON_NAME;
      ci->idInfo = IDS_DIALOG_TITLE;
      ci->lData = 0;
      } return 1;

   case CPL_DBLCLK: {
      RunNTPTimeControlPanel(hinst, hwndCPL);
      } return 1;

   }
   return 0;
}
