#ifndef ntp_service_h
#define ntp_service_h

// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include "wrap-service.h"
#include "wrap-critical.h"

class LogStuff;
class NtpLongFixed;
class NtpEvents;

class NtpService : public WrapService {
public:

   NtpService(HINSTANCE init_hInstance);

   virtual
   ~NtpService();

   virtual void
   Run();

   virtual bool
   OnInit();

   virtual void
   OnStop();

   virtual void
   OnInterrogate();

   virtual void
   OnPause();

   virtual void
   OnContinue();

   virtual void
   OnShutdown();

   virtual bool
   OnUserControl(DWORD user_opcode);

   virtual LogStuff *
   GetLog();

   virtual LPCTSTR
   GetDependencyList();

   virtual LPTSTR
   GetDescriptiveText();

   void
   SetUseRasDependency(bool newdep);

   static char * ntpversion;

private:

   bool
   IsRasActive();

   void
   TryOnce();

   void
   LogTimeDiff(const char * msg, NtpLongFixed & t);

   void
   ResetRetry();

   enum WaitReasons {
      STOP_EVENT,
      POKE_EVENT,
      TIMEOUT_EVENT
   };

   WaitReasons
   WaitForSomething(DWORD milliseconds);

   bool         use_ras_dependency;
   bool         wait_for_ras;
   bool         ras_dll_loaded;
   HINSTANCE    ras_dll_handle;
   FARPROC      ras_enum_connections_ptr;
   FARPROC      ras_get_connection_status_ptr;
   FARPROC      ras_get_error_string_ptr;
   bool         run_once;
   bool         write_log;
   volatile bool shutting_down;
   DWORD        ras_poll_milliseconds;
   DWORD        ntp_poll_milliseconds;
   char **      ntp_server_names;
   int          ntp_server_count;
   int          which_server;
   unsigned short ntp_server_port;
   unsigned char ntp_protocol_version;
   char *       log_directory_name;
   DWORD        max_adjust_seconds;
   DWORD        log_file_count;
   bool         last_write_log;
   char *       last_log_directory_name;

   HANDLE       stop_event;

   bool         allow_poke;
   bool         allow_monitor;
   NtpEvents *  ntp_events;

   DWORD        retry_init_milliseconds;
   DWORD        retry_max_milliseconds;
   DWORD        retry_milliseconds;
   DWORD        retry_add;
   DWORD        retry_mult;
   DWORD        retry_div;
   DWORD        retry_count;
   DWORD        retry_max_count;
   bool         kissed_off;

   DWORD        startup_delay_milliseconds;
   DWORD        ras_delay_milliseconds;
   DWORD        shutdown_delay_milliseconds;

   DWORD        ntp_reply_port;
   bool         disable_packet_error_checks;

   DeclareCriticalSection run_lock;
};

#endif // ntp_service_h
