#ifndef regstuff_h
#define regstuff_h

// Copyright (C) Tom Horsley, 1997. All rights reserved.

// Classes to simplify reading values from the registry.

#include <windows.h>

class RegistryKeyBase {
public:

      // Constructors open an existing key for reading values or
      // throw exception.
      //
   RegistryKeyBase(HKEY parent_key, LPCTSTR key_name);

   RegistryKeyBase(RegistryKeyBase & parent_key, LPCTSTR key_name);

   virtual
   ~RegistryKeyBase();

      // Get a value in REG_DWORD format or throw exception
      //
   DWORD
   GetRegDwordValue(const char * valname);

      // Get a value in REG_SZ format or throw exception
      //
   char *
   GetRegSzValue(const char * valname);

      // Get a value in REG_SZ or REG_EXPAND_SZ format or throw exception.
      // If format was REG_EXPAND_SZ, then expand the registers and return
      // the expanded value.
      //
   char *
   GetRegExpandSzValue(const char * valname);

      // Ask about a value - returns true if all ok (and fills in format
      // and length parameters), otherwise returns false.
      //
   bool
   QueryValue(const char * valname, DWORD & val_format, DWORD & val_length);

protected:

   RegistryKeyBase(HKEY init_key);

   void
   CommonInit(HKEY parent_key, LPCTSTR key_name);

   HKEY my_key;

private:

   virtual void
   Cleanup() = 0;
};

class RegistryPredefinedKey : public RegistryKeyBase {
protected:

   RegistryPredefinedKey(HKEY init_key);

public:

   virtual
   ~RegistryPredefinedKey();

private:

   virtual void
   Cleanup();
};

class RegistryClassesRoot : public RegistryPredefinedKey {
public:

   RegistryClassesRoot();

   virtual
   ~RegistryClassesRoot();
};

class RegistryCurrentUser : public RegistryPredefinedKey {
public:

   RegistryCurrentUser();

   virtual
   ~RegistryCurrentUser();
};

class RegistryLocalMachine : public RegistryPredefinedKey {
public:

   RegistryLocalMachine();

   virtual
   ~RegistryLocalMachine();
};

class RegistryUsers : public RegistryPredefinedKey {
public:

   RegistryUsers();

   virtual
   ~RegistryUsers();
};

class RegistryKey : public RegistryKeyBase {
public:

   RegistryKey(HKEY parent_key, LPCTSTR key_name);

   RegistryKey(RegistryKeyBase & parent_key, LPCTSTR key_name);

   virtual
   ~RegistryKey();

private:

   virtual void
   Cleanup();
};

#endif // regstuff_h
