#ifndef wrap_critical_h
#define wrap_critical_h

// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <windows.h>

// Use this class to make sure a critical section is entered and left
// correctly while the object of this type is in scope.
//
class WrapCriticalSection {
public:

   WrapCriticalSection(CRITICAL_SECTION * init_cp)
      : cp(init_cp) { EnterCriticalSection(cp); };

   WrapCriticalSection(CRITICAL_SECTION & init_cp)
      : cp(&init_cp) { EnterCriticalSection(cp); };

   ~WrapCriticalSection()
      { LeaveCriticalSection(cp); };

private:

   CRITICAL_SECTION * cp;
};

// Use this class to declare a CRITICAL_SECTION object which is properly
// initialized and cleaned up.
//
class DeclareCriticalSection {
public:

   DeclareCriticalSection()
      { InitializeCriticalSection(&cs); };

   ~DeclareCriticalSection()
      { DeleteCriticalSection(&cs); };

   operator CRITICAL_SECTION * ()
      {return &cs;};

private:

   CRITICAL_SECTION cs;
};

#endif // wrap_critical_h
