#ifndef base_exception_h
#define base_exception_h

// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <exception>
#include <string>
#include <stdarg.h>

class FunctionTracer;

// A BaseException defines some additional member functions over and above
// the standard exception class. It is possible for a BaseException to
// accumulate a string of reasons that describe all the calamities which
// occurred at each stage of a cascade of exceptions :-).
//
class BaseException : public std::exception {
public:

   BaseException() throw();

   BaseException(const BaseException& rhs) throw();

   BaseException&
   operator=(const BaseException& rhs) throw();

   virtual 
   ~BaseException() throw();

   void
   Printf(const char * fmt, ...);

   void
   VPrintf(const char * fmt, va_list vl);

   void
   AddCalledFromInfo(const FunctionTracer & tm);

   void
   AddThrowLocation(const char * file, int line);

   const std::string &
   GetDescription();

   virtual const char *
   what() const throw();

protected:

   void
   ConditionalNewline();

   std::string description;
};

#endif // base_exception_h
